import django_filters
import netaddr
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.utils.translation import gettext as _
from netaddr import AddrFormatError

from ipam.models import IPAddress
from netbox.filtersets import NetBoxModelFilterSet, BaseFilterSet
from dcim.models import Device
from netbox_config_backup import models
from netbox_config_backup.choices import FileTypeChoices
from utilities.filters import MultiValueCharFilter


class BackupFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )
    device = django_filters.ModelMultipleChoiceFilter(
        field_name='device__name',
        queryset=Device.objects.all(),
        to_field_name='name',
        label='Device (name)',
    )
    device_id = django_filters.ModelMultipleChoiceFilter(
        field_name='device',
        queryset=Device.objects.all(),
        label='Device (ID)',
    )
    ip = django_filters.CharFilter(
        method='filter_address',
        label=_('Address'),
    )

    class Meta:
        model = models.Backup
        fields = ['id', 'name', 'ip']

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = (
            Q(name__icontains=value) |
            Q(device__name__icontains=value) |
            Q(device__primary_ip4__address__contains=value.strip()) |
            Q(device__primary_ip6__address__contains=value.strip()) |
            Q(ip__address__contains=value.strip())
        )

        try:
            prefix = str(netaddr.IPNetwork(value.strip()).cidr)
            qs_filter |= Q(device__primary_ip4__address__net_host_contained=prefix)
            qs_filter |= Q(device__primary_ip6__address__net_host_contained=prefix)
            qs_filter |= Q(ip__address__net_host_contained=prefix)

        except (AddrFormatError, ValueError):
            pass

        return queryset.filter(qs_filter)

    def filter_address(self, queryset, name, value):
        try:
            return queryset.filter(ip__address__net_host_contained=value)
        except ValidationError:
            return queryset.none()


class BackupsFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )
    type = django_filters.MultipleChoiceFilter(
        field_name='file__type',
        choices=FileTypeChoices,
        null_value=None
    )

    class Meta:
        model = models.BackupCommitTreeChange
        fields = ['id', 'file']

    def search(self, queryset, name, value):
        print('Search')
        if not value.strip():
            return queryset
        print(value)
        qs_filter = (
            Q(file__type=value) |
            Q(file__type__startswith=value)
        )
        return queryset.filter(qs_filter)



from extras.plugins import PluginConfig
from importlib.metadata import metadata

metadata = metadata('netbox_config_backup')


class NetboxConfigBackup(PluginConfig):
    name = metadata.get('Name').replace('-', '_')
    verbose_name = metadata.get('Summary')
    description = metadata.get('Description')
    version = metadata.get('Version')
    author = metadata.get('Author')
    author_email = metadata.get('Author-email')
    base_url = 'configbackup'
    min_version = '3.5.8'
    max_version = '3.7.99'
    required_settings = [
        'repository',
        'committer',
        'author',
    ]
    default_settings = {
        # Frequency in seconds
        'frequency': 3600,
    }
    queues = [
        'jobs'
    ]


config = NetboxConfigBackup



import uuid

from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils import timezone

from netbox_config_backup.models import BackupJob
from netbox_config_backup.tasks import backup_job
from netbox_config_backup.utils import remove_queued
from netbox_config_backup.utils.rq import can_backup


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('--time', dest='time', help="time")
        parser.add_argument('--device', dest='device', help="Device Name")

    def run_backup(self, backup):
        if can_backup(backup):
            backupjob = backup.jobs.filter(backup__device=backup.device).last()
            if backupjob is None:
                backupjob = BackupJob.objects.create(
                    backup=backup,
                    scheduled=timezone.now(),
                    uuid=uuid.uuid4()
                )
            backup_job(backupjob.pk)
        remove_queued(backup)

    def handle(self, *args, **options):
        from netbox_config_backup.models import Backup
        if options['device']:
            print(f'Running:{options.get("device")}| ')
            backup = Backup.objects.filter(device__name=options['device']).first()
            if backup:
                self.run_backup(backup)
            else:
                raise Exception('Device not found')
        else:
            for backup in Backup.objects.all():
                self.run_backup(backup)




# Generated by Django 3.2.8 on 2021-11-23 02:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_config_backup', '0002_git_models'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='backup',
            name='created',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='custom_field_data',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='last_updated',
        ),
        migrations.RemoveField(
            model_name='backup',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='backupcommit',
            name='created',
        ),
        migrations.RemoveField(
            model_name='backupcommit',
            name='custom_field_data',
        ),
        migrations.RemoveField(
            model_name='backupcommit',
            name='last_updated',
        ),
        migrations.RemoveField(
            model_name='backupcommit',
            name='tags',
        ),
    ]



import uuid

from django.core.management.base import BaseCommand
from django.db import transaction
from django.utils import timezone

from netbox_config_backup.models import BackupJob
from netbox_config_backup.tasks import backup_job
from netbox_config_backup.utils import remove_queued
from netbox_config_backup.utils.rq import can_backup


class Command(BaseCommand):

    def handle(self, *args, **options):
        from netbox_config_backup.models import Backup
        print(f'Backup Name\t\tDevice Name\t\tIP')
        for backup in Backup.objects.filter(device__isnull=False):
            if backup.ip:
                ip = backup.ip
            else:
                ip = backup.device.primary_ip

            name = f'{backup.name}'
            if len(backup.name) > 15:
                name = f'{name}\t'
            elif len(backup.name) > 7:
                name = f'{name}\t\t'
            else:
                name = f'{name}\t\t\t'

            device_name = f'{backup.device.name}'
            if len(backup.device.name) > 15:
                device_name = f'{device_name}\t'
            elif len(backup.device.name) > 7:
                device_name = f'{device_name}\t\t'
            else:
                device_name = f'{device_name}\t\t\t'

            print(f'{name}{device_name}{ip}')



from extras.plugins import PluginMenuItem, PluginMenuButton, PluginMenu
from utilities.choices import ButtonColorChoices

assigned = PluginMenuItem(
    link='plugins:netbox_config_backup:backup_list',
    link_text='Devices',
    permissions=['netbox_config_backup.view_backups'],
    buttons=[
        PluginMenuButton(
            link="plugins:netbox_config_backup:backup_add",
            title="Add",
            icon_class="mdi mdi-plus",
            color=ButtonColorChoices.GREEN,
        ),
    ]
)

unassigned = PluginMenuItem(
    link='plugins:netbox_config_backup:unassignedbackup_list',
    link_text='Unassigned Backups',
    permissions=['netbox_config_backup.view_backups'],
    buttons=[]
)

menu = PluginMenu(
    label="Configuration Backup",
    groups=(
        ('Backup Jobs', (assigned, unassigned)),
    )
)



import logging
import datetime
import uuid as uuid

from django.db import models
from django.db.models import Q
from django.urls import reverse

from django_rq import get_queue

from dcim.models import Device
from extras.choices import JobResultStatusChoices
from netbox.models import NetBoxModel

from netbox_config_backup.choices import StatusChoices
from netbox_config_backup.helpers import get_repository_dir

from netbox_config_backup.utils.rq import remove_queued
from ..querysets import BackupQuerySet
from ..utils import Differ

logger = logging.getLogger(f"netbox_config_backup")


class Backup(NetBoxModel):
    name = models.CharField(max_length=255, unique=True)
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    status = models.CharField(
        max_length=50,
        choices=StatusChoices,
        default=StatusChoices.STATUS_ACTIVE
    )
    device = models.ForeignKey(
        to=Device,
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )
    ip = models.ForeignKey(
        to='ipam.IPAddress',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    objects = BackupQuerySet.as_manager()

    class Meta:
        ordering = ['name']

    @property
    def backup_count(self):
        return self.changes.count()

    def get_absolute_url(self):
        return reverse('plugins:netbox_config_backup:backup', args=[self.pk])

    def __str__(self):
        return self.name

    def delete(self, *args, **kwargs):
        queue = get_queue('netbox_config_backup.jobs')
        remove_queued(self)

        super().delete(*args, **kwargs)

    def enqueue_if_needed(self):
        from netbox_config_backup.utils.rq import enqueue_if_needed
        return enqueue_if_needed(self)

    def requeue(self):
        self.jobs.filter(
            ~Q(status=JobResultStatusChoices.STATUS_COMPLETED) &
            ~Q(status=JobResultStatusChoices.STATUS_FAILED) &
            ~Q(status=JobResultStatusChoices.STATUS_ERRORED)
        ).update(
            status=JobResultStatusChoices.STATUS_FAILED
        )
        remove_queued(self)
        self.enqueue_if_needed()

    def get_config(self, index='HEAD'):
        from netbox_config_backup.git import repository
        running = repository.read(f'{self.uuid}.running')
        startup = repository.read(f'{self.uuid}.startup')

        return {'running': running if running is not None else '', 'startup': startup if startup is not None else ''}

    def set_config(self, configs, files=('running', 'startup'), pk=None):
        from netbox_config_backup.models.repository import (BackupCommit, BackupObject, BackupFile,
                                                            BackupCommitTreeChange)
        from netbox_config_backup.git import repository
        LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

        stored_configs = self.get_config()
        changes = False
        for file in files:
            #logger.debug(f'[{pk}] Getting existing config for {file}')
            stored = stored_configs.get(file) if stored_configs.get(file) is not None else ''
            #logger.debug(f'[{pk}] Getting new config')
            current = configs.get(file) if configs.get(file) is not None else ''
            #logger.debug(f'[{pk}] Starting diff for {file}')
            if Differ(stored, current).is_diff():
                changes = True
                output = repository.write(f'{self.uuid}.{file}', current)
            #logger.debug(f'[{pk}] Finished diff for {file}')
        if not changes:
            return None

        #logger.debug(f'[{pk}] Commiting files')
        commit = repository.commit(f'Backup of {self.device.name} for backup {self.name}')

        #logger.debug(f'[{pk}] Getting repository log')
        log = repository.log(index=commit, depth=1)[0]

        #logger.debug(f'[{pk}] Saving commit to DB')
        bc = BackupCommit.objects.filter(sha=commit)
        time = log.get('time', datetime.datetime.now()).replace(tzinfo=LOCAL_TIMEZONE)
        if bc.count() > 0:
            #logger.debug(f'[{pk}] Error committing')
            raise Exception('Commit already exists for this backup and sha value')
        else:
            #logger.debug(f'[{pk}] Saving commit')
            bc = BackupCommit(sha=commit, time=time)
            logger.info(f'{self}: {commit}:{bc.time}')
            bc.save()

        for change in log.get('changes', []):
            #logger.debug(f'[{pk}] Adding backup tree changes')
            backupfile = None
            change_data = {}
            for key in ['old', 'new']:
                sha = change.get(key, {}).get('sha', None)
                file = change.get(key, {}).get('path', None)
                if sha is not None and file is not None:
                    uuid, type = file.split('.')
                    try:
                        object = BackupObject.objects.get(sha=sha)
                    except BackupObject.DoesNotExist:
                        object = BackupObject.objects.create(sha=sha)
                    try:
                        backupfile = BackupFile.objects.get(backup=self, type=type)
                    except BackupFile.DoesNotExist:
                        backupfile = BackupFile.objects.create(backup=self, type=type)
                    change_data[key] = object

            bctc = BackupCommitTreeChange.objects.filter(
                backup=self,
                file=backupfile,
                commit=bc,
                type=change.get('type', None),
                old=change_data.get('old', None),
                new=change_data.get('new', None)
            )
            if bctc.count() > 0:
                bctc = bctc.first()
            elif bctc.count() == 0:
                bctc = BackupCommitTreeChange(
                    backup=self,
                    file=backupfile,
                    commit=bc,
                    type=change.get('type', None),
                    old=change_data.get('old', None),
                    new=change_data.get('new', None)
                )
                bctc.save()
                #logger.debug(f'[{pk}] Tree saved')
        #logger.debug(f'[{pk}] Get config saved')
        return commit

    @classmethod
    def get_repository_dir(cls):
        return get_repository_dir()


from django.db import models

from extras.choices import JobResultStatusChoices
from utilities.querysets import RestrictedQuerySet



class BackupQuerySet(RestrictedQuerySet):
    def default_annotate(self):
        from netbox_config_backup.models import BackupJob, BackupCommitTreeChange

        return self.prefetch_related('jobs').annotate(
        last_backup=models.Subquery(
            BackupJob.objects.filter(backup=models.OuterRef('id'), status=JobResultStatusChoices.STATUS_COMPLETED).order_by('completed').values('completed')[:1]
        ),
        next_attempt=models.Subquery(
            BackupJob.objects.filter(backup=models.OuterRef('id'), status__in=['pending', 'running']).order_by('scheduled').values('scheduled')[:1]
        ),
        last_change=models.Subquery(
            BackupCommitTreeChange.objects.filter(backup=models.OuterRef('id')).values('commit__time')[:1]
        )
    )


import os

from netbox import settings


def get_repository_dir():
    repository = settings.PLUGINS_CONFIG.get("netbox_config_backup", {}).get("repository")
    if repository == os.path.abspath(repository) or repository == (os.path.abspath(repository) + os.path.sep):
        return repository

    return f'{os.path.dirname(os.path.dirname(settings.BASE_DIR))}{os.path.sep}{repository}'


from .backups import get_backup_tables
from .git import Differ
from .rq import enqueue_if_needed, enqueue, needs_enqueue, remove_queued, remove_orphaned, is_queued, is_running

__all__ = (
    'get_backup_tables',
    'Differ',
    'enqueue',
    'enqueue_if_needed',
    'needs_enqueue',
    'remove_queued',
    'remove_orphaned',
    'is_running',
    'is_queued',
)


# Generated by Django 4.1.7 on 2023-04-05 17:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_config_backup', '0011_alter_backup_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='backup',
            name='status',
            field=models.CharField(default='active', max_length=50),
        ),
    ]



# Generated by Django 3.2.9 on 2022-02-28 20:47

from django.db import migrations, models
import django.db.models.deletion


def migrate_from_bo_to_bctc(apps, schema_editor):
    print('')
    print('!!!WARNING!!!')
    print('Once this migration operation is complete, it is not recommended to revert it')

    BackupCommitTreeChange = apps.get_model('netbox_config_backup', 'BackupCommitTreeChange')
    BackupFile = apps.get_model('netbox_config_backup', 'BackupFile')
    bctcs = BackupCommitTreeChange.objects.all()
    for bctc in bctcs:
        if bctc.new:
            uuid, ftype = bctc.new.file.split('.')
        else:
            uuid, ftype = bctc.old.file.split('.')

        try:
            bf = BackupFile.objects.get(backup=bctc.commit.backup, type=ftype)
        except BackupFile.DoesNotExist:
            bf = BackupFile.objects.create(backup=bctc.commit.backup, type=ftype)

        bctc.backup = bctc.commit.backup
        bctc.file = bf

        bctc.save()


def migrate_from_bctc_to_bo(apps, schema_editor):
    print('')
    print('!!!!WARNING!!!')
    print('This migration operation may fail and leave your netbox_config_backup database in an onconsistent state')

    BackupCommitTreeChange = apps.get_model('netbox_config_backup', 'BackupCommitTreeChange')
    bctcs = BackupCommitTreeChange.objects.all()
    for bctc in bctcs:
        bctc.refresh_from_db()
        bctc.commit.refresh_from_db()

        bctc.commit.backup = bctc.backup
        bctc.commit.save()

        if bctc.new:
            bctc.new.refresh_from_db()
            bctc.new.type = f'{bctc.backup.uuid}.{bctc.file.type}'
            bctc.new.save()
        if bctc.old:
            bctc.old.refresh_from_db()
            bctc.old.file = f'{bctc.backup.uuid}.{bctc.file.type}'
            bctc.old.save()


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_config_backup', '0008_backupjob_scheduled_nullable'),
    ]

    operations = [
        migrations.AlterField(
            model_name='backupobject',
            name='sha',
            field=models.CharField(max_length=64, unique=True),
        ),
        migrations.AlterUniqueTogether(
            name='backupobject',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='backupcommit',
            name='time',
            field=models.DateTimeField(),
        ),
        migrations.CreateModel(
            name='BackupFile',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('type', models.CharField(max_length=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='backupfile',
            name='backup',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='files', to='netbox_config_backup.backup'),
        ),
        migrations.AlterUniqueTogether(
            name='backupfile',
            unique_together={('backup', 'type')},
        ),
        migrations.AddField(
            model_name='backupcommittreechange',
            name='backup',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='changes', to='netbox_config_backup.backup'),
        ),
        migrations.AddField(
            model_name='backupcommittreechange',
            name='file',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='changes', to='netbox_config_backup.backupfile'),
        ),
        migrations.AlterField(
            model_name='backupcommittreechange',
            name='new',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='changes', to='netbox_config_backup.backupobject'),
        ),
        migrations.RunPython(migrate_from_bo_to_bctc, migrate_from_bctc_to_bo),
        migrations.AlterField(
            model_name='backupcommittreechange',
            name='backup',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='changes', to='netbox_config_backup.backup'),
        ),
        migrations.AlterField(
            model_name='backupcommittreechange',
            name='file',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='changes', to='netbox_config_backup.backupfile'),
        ),
        migrations.RemoveField(
            model_name='backupobject',
            name='file',
        ),
        migrations.RemoveConstraint(
            model_name='backupcommit',
            name='backup_and_sha_not_null',
        ),
        migrations.RemoveField(
            model_name='backupcommit',
            name='backup',
        ),
    ]



import logging
import sys
import traceback
from datetime import timedelta

from django.utils import timezone
from django_rq import job
from netmiko import NetmikoAuthenticationException, NetmikoTimeoutException

from extras.choices import JobResultStatusChoices
from netbox import settings
from netbox.api.exceptions import ServiceUnavailable
from netbox.config import get_config
from netbox_config_backup.models import Backup, BackupJob, BackupCommit
from netbox_config_backup.utils.rq import can_backup


def get_logger():
    # Setup logging to Stdout
    formatter = logging.Formatter(f'[%(asctime)s][%(levelname)s] - %(message)s')
    stdouthandler = logging.StreamHandler(sys.stdout)
    stdouthandler.setLevel(logging.DEBUG)
    stdouthandler.setFormatter(formatter)
    logger = logging.getLogger(f"netbox_config_backup")
    logger.addHandler(stdouthandler)

    return logger


def napalm_init(device, ip=None, extra_args={}):
    from netbox import settings
    username = settings.PLUGINS_CONFIG.get('netbox_napalm_plugin', {}).get('NAPALM_USERNAME', None)
    password = settings.PLUGINS_CONFIG.get('netbox_napalm_plugin', {}).get('NAPALM_PASSWORD', None)
    timeout = settings.PLUGINS_CONFIG.get('netbox_napalm_plugin', {}).get('NAPALM_TIMEOUT', None)
    optional_args = settings.PLUGINS_CONFIG.get('netbox_napalm_plugin', {}).get('NAPALM_ARGS', []).copy()

    if device and device.platform and device.platform.napalm.napalm_args is not None:
        optional_args.update(device.platform.napalm.napalm_args)
    if extra_args != {}:
        optional_args.update(extra_args)

    # Check for primary IP address from NetBox object
    if ip is not None:
        host = str(ip.address.ip)
    elif device.primary_ip and device.primary_ip is not None:
        host = str(device.primary_ip.address.ip)
    else:
        raise ServiceUnavailable(
            "This device does not have a primary IP address"
        )

    # Check that NAPALM is installed
    try:
        import napalm
        from napalm.base.exceptions import ModuleImportError
    except ModuleNotFoundError as e:
        if getattr(e, 'name') == 'napalm':
            raise ServiceUnavailable("NAPALM is not installed. Please see the documentation for instructions.")
        raise e

    # Validate the configured driver
    try:
        driver = napalm.get_network_driver(device.platform.napalm.napalm_driver)
    except ModuleImportError:
        raise ServiceUnavailable("NAPALM driver for platform {} not found: {}.".format(
            device.platform, device.platform.napalm.napalm_driver
        ))

    # Connect to the device
    d = driver(
        hostname=host,
        username=username,
        password=password,
        timeout=timeout,
        optional_args=optional_args
    )
    try:
        d.open()
    except Exception as e:
        if isinstance(e, NetmikoAuthenticationException):
            logger.info('Authentication error')
        elif isinstance(e, NetmikoTimeoutException):
            logger.info('Connection error')
        raise ServiceUnavailable("Error connecting to the device at {}: {}".format(host, e))

    return d


def backup_config(backup, pk=None):
    commit = None
    if backup.device:
        ip = backup.ip if backup.ip is not None else backup.device.primary_ip
    else:
        ip = None
    if not can_backup(backup):
        raise Exception(f'Cannot backup {backup}')

    if backup.device is not None and ip is not None:
        logger.info(f'{backup}: Backup started')
        #logger.debug(f'[{pk}] Connecting')
        d = napalm_init(backup.device, ip)
        #logger.debug(f'[{pk}] Finished Connection')

        #logger.debug(f'[{pk}] Getting config')
        configs = d.get_config()
        #logger.debug(f'[{pk}] Finished config get')

        #logger.debug(f'[{pk}] Setting config')
        commit = backup.set_config(configs, pk=pk)
        #logger.debug(f'[{pk}] Finished config set')

        d.close()
        logger.info(f'{backup}: Backup complete')
    else:
        logger.info(f'{backup}: No IP set')

    return commit


def backup_job(pk):
    import netmiko
    try:
        job_result = BackupJob.objects.get(pk=pk)
    except BackupJob.DoesNotExist:
        logger.error(f'Cannot locate job (Id: {pk}) in DB')
        raise Exception(f'Cannot locate job (Id: {pk}) in DB')
    backup = job_result.backup

    if not can_backup(backup):
        logger.warning(f'Cannot backup due to additional factors')
        return 1
    delay = timedelta(seconds=settings.PLUGINS_CONFIG.get('netbox_config_backup', {}).get('frequency'))

    job_result.started = timezone.now()
    job_result.status = JobResultStatusChoices.STATUS_RUNNING
    job_result.save()
    try:
        #logger.debug(f'[{pk}] Starting backup')
        commit = backup_config(backup, pk=pk)
        #logger.debug(f'[{pk}] Finished backup')

        job_result.set_status(JobResultStatusChoices.STATUS_COMPLETED)
        job_result.data = {'commit': f'{commit}' if commit is not None else ''}
        job_result.set_status(JobResultStatusChoices.STATUS_COMPLETED)
        # Enqueue next job if one doesn't exist
        try:
            #logger.debug(f'[{pk}] Starting Enqueue')
            BackupJob.objects.filter(
                backup=backup
            ).exclude(
                status__in=JobResultStatusChoices.TERMINAL_STATE_CHOICES
            ).update(status=JobResultStatusChoices.STATUS_FAILED)
            BackupJob.enqueue_if_needed(backup, delay=delay, job_id=job_result.job_id)
            #logger.debug(f'[{pk}] Finished Enqueue')
        except Exception as e:
            logger.error(f'Job Enqueue after completion failed for job: {backup}')
            logger.error(f'\tException: {e}')
    except netmiko.exceptions.ReadTimeout as e:
        BackupJob.enqueue_if_needed(backup, delay=delay, job_id=job_result.job_id)
        logger.warning(f'Netmiko read timeout on job: {backup}')
    except ServiceUnavailable as e:
        logger.info(f'Napalm service read failure on job: {backup} ({e})')
        BackupJob.enqueue_if_needed(backup, delay=delay, job_id=job_result.job_id)
    except Exception as e:
        logger.error(f'Uncaught Exception on job: {backup}')
        logger.error(e)
        logger.warning(traceback.format_exc())
        job_result.set_status(JobResultStatusChoices.STATUS_ERRORED)
        BackupJob.enqueue_if_needed(backup, delay=delay, job_id=job_result.job_id)

    #logger.debug(f'[{pk}] Saving result')
    job_result.save()


logger = get_logger()
